#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "../utils/exhaust/all.h"
#include "../utils/utils.h"
#include "fitness.h"

void print_koth_status(int * scores, warrior_t * w, int anz);

void fitness_koth(warrior_t * w, int anz, int update) {
    static warrior_t * koth_warriors=0;
    static int * koth_scores=0;
    static int worst_nr=0;
    int veteran_nr=0;
    static unsigned long worst_score=-1; /* highest possible number */
    int old_nr=0;
    unsigned long old_score=0;
    unsigned long score=0;
    int x, y, i;
    int fights;

    /* init at first start */
    if (koth_warriors==0) {

        /* allocate mem */
        koth_warriors = malloc(sizeof(warrior_t)*(globals.koth_size+1));
        koth_scores = malloc(sizeof(int)*(globals.koth_size+1)*(globals.koth_size+1));

        /* load/create warriors */
        create_warriors(koth_warriors, globals.koth_size+1, globals.startlength, "koth");

        /* recalc scores with no self-fights */
        printf("calculating hill-scores...\n");
        koth_warriors[0].no=0;
        for (y=1; y<globals.koth_size+1; y++) {
            koth_warriors[y].no=-1; /* Age of the warrior */
            for (x=0; x<y; x++) {
                printf("fight: %3d vs. %3d\n", y, x);
                koth_warriors[x].score=0;
                koth_warriors[y].score=0;
                mars(&koth_warriors[x], &koth_warriors[y], globals.koth_battles);
                koth_scores[x + y*(globals.koth_size+1)] = koth_warriors[y].score;
                koth_scores[y + x*(globals.koth_size+1)] = koth_warriors[x].score;
            }
        }
        /* set nr of worst warrior */
        for (y=0; y<globals.koth_size + 1; y++) {
            score = 0;
            for (x=0; x<globals.koth_size + 1; x++) {
                if (x == y) continue; /* no self-fights */
                score += koth_scores[x + y*(globals.koth_size+1)];
            }               
            if (score < worst_score) {
                worst_nr = y;
                worst_score = score;
            }
        }
        printf("hill-scores ready.\n");
    }

    /* update=0: no warrior is replaced. */
    if (update == 0) {
        /* every warrior fights number_of_battles times */
        fights = globals.number_of_battles / (globals.koth_size+1);
        if (fights < 1) fights=1;
        
        for (i=0; i<anz; i++) {
            w[i].score=0;
            w[i].fights=0;
            for (y=0; y<globals.koth_size+1; y++) {
                if (y == worst_nr) continue;
                mars(&w[i], &koth_warriors[y], fights);
            }
        }
        /* print scores */
    }
    else {
        fights = globals.koth_battles;
        
        for (i=0; i<anz; i++) {
            /* worst warrior is always replaced */
            koth_warriors[worst_nr] = w[i];
            koth_warriors[worst_nr].no=0;
            for (x=0; x<globals.koth_size+1; x++) {
                if (x == worst_nr) continue; /* no self-fights */
                koth_warriors[x].score=0;
                koth_warriors[x].no++;
                koth_warriors[worst_nr].score=0;
                
                mars(&koth_warriors[worst_nr], &koth_warriors[x], fights);
                koth_scores[x + worst_nr*(globals.koth_size+1)] = koth_warriors[worst_nr].score;
                koth_scores[worst_nr + x*(globals.koth_size+1)] = koth_warriors[x].score;
            }

            /* find worst warrior + oldest warrior */
            old_nr = worst_nr;
            veteran_nr = 0;
            worst_score=-1; /* must be unsigned! */
            
            for (y=0; y<globals.koth_size + 1; y++) {
                score = 0;
                for (x=0; x<globals.koth_size + 1; x++) {
                    if (x == y) continue; /* no self-fights */
                    score += koth_scores[x + y*(globals.koth_size+1)];
                }               
                if (score < worst_score) {
                    worst_nr = y;
                    worst_score = score;
                }
                if (y == old_nr) {
                    old_score = score;
                }
                if (koth_warriors[y].no > koth_warriors[veteran_nr].no) {
                    veteran_nr = y;
                }
            }
            
            if (old_nr != worst_nr) {
                printf("New warrior on hill!  nr %d\n", old_nr);
                save_warrior(&koth_warriors[old_nr], old_nr, globals.all_battles, "koth");
                print_koth_status(koth_scores, koth_warriors, globals.koth_size+1);
            }
            if (probability(globals.reintroduce_veteran_rate)) {
                /* dumb: replaces any warrior from the soup */
                printf("reintroducing veteran nr. %d\n", veteran_nr);
                w[random_number(anz)] = koth_warriors[veteran_nr];
            }
        }
    }
}

void print_koth_status(int * scores, warrior_t * w, int anz) {
    FILE * output;
    char file_name[256];
    int x, y, i, j;
    double score;
    warrior_t ** list=0;

    int max_color=255;
    int color;

    /* open status-file */
    sprintf(file_name, "%s%sstatus.htm", "koth", globals.path_symbol);
    if (!(output = fopen(file_name, "w"))) {
        printf("Oh... I can not create the outputfile '%s'. No, don't panic.\n", file_name);
        printf("dont panic. Dont Panic! DONT PANIC!!! PAAANNNICCCC!\n");
        for (x=0; x<100; x++) printf(" PAAAAANIIIIIIIICCCC!!!");
        return;
    }

    /* calculate+set warrior scores */
    for (y=0; y<anz; y++) {
        w[y].score=0;
        for (x=0; x<anz; x++) {
            if (x == y) continue; /* no self-fights */
            w[y].score += scores[x + y*anz];
        }
        w[y].fights=(anz-1)*2*globals.koth_battles;
    }

    /* get sorted list of warriors */
    list = sorted_list(w, anz);
    
    /* print header */
    fprintf(output, "<html><body bgcolor=\"#ffffff\">\n");
    fprintf(output, "<h1>yace - king of the hill</h1>\n");
    /* fprintf(output, "<a href=\"../index.html\">Main</a> / koth<p>"); */
    fprintf(output, "<center>\n");
    fprintf(output, "<table cellpadding=\"2\" cellspacing=\"1\">\n<tr>");
    fprintf(output, "<td>&nbsp</td>");
    for (x=0; x<anz; x++)
        fprintf(output, "<td align=\"right\" bgcolor=\"#a4ddfe\"><a href=\"%08d.red\">%d</a></td>", list[anz-1-x] - &w[0], list[anz-1-x] - &w[0]);
    fprintf(output, "<td bgcolor=\"#a4ddfe\">Score</td><td bgcolor=\"#a4ddfe\">Age</td></tr>\n");
    for (i=0; i<anz; i++) {
        y = list[anz-1-i] - &w[0];

        /* <hr> for last warrior */
        if (i == anz-1) {
            fprintf(output, "<tr><td colspan=\"%d\"><hr noshade size=\"1\"></td></tr>\n", anz+3);
        }
        
        /* table row starts here */
        fprintf(output, "<tr><td align=\"right\" bgcolor=\"#a4ddfe\"><a href=\"%08d.red\">%d</a></td>", y, y);
        score=0;
        for (j=0; j<anz; j++) {
            x = list[anz-1-j] - &w[0];
            /* table cell starts here */
            if (x == y) { /* no self-fights */
                fprintf(output, "<td align=\"right\" bgcolor=\"#%02x%02x%02x\">&nbsp</td>", 255-max_color/2, 255-max_color/2, 255-max_color/2);
                continue;
            }
            color = (int)(255.0 - 1.0*max_color/300*scores[x + y*anz]*100.0/(globals.koth_battles*2));
            fprintf(output, "<td align=\"right\" bgcolor=\"#%02x%02x%02x\">", color, color, color);
            if (color <= 86) fprintf(output, "<font color=\"#fffff\">");
            fprintf(output, "%3.0f</td>", scores[x + y*anz]*100.0/(globals.koth_battles*2));
            if (color <= 86) fprintf(output, "</font>");
            score += scores[x + y*anz];
        }
        /* table row ends here */
        fprintf(output, "<td align=\"right\" bgcolor=\"dfdfdf\">%5.1f</td><td align=\"right\" bgcolor=\"dfdfdf\">%d</td></tr>\n",
                score*100.0/((anz-1)*2*globals.koth_battles),
                w[y].no);
    }
    
    fprintf(output, "</table></center>");
    fprintf(output, "<p>(c) by Martin Ankerl 2000");
    fprintf(output, "</body></html>");
    fclose(output);
}
