#include <math.h> /* sqrt */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "../utils/exhaust/all.h"
#include "../utils/utils.h"
#include "fitness.h"

void koth_init(warrior_t ** koth_warriors, double ** koth_score, long ksize);
void print_kothsimi_status(double * scores, warrior_t * w, int anz);
int koth_worst_nr(double * koth_score, int size);
int get_most_similar_warrior(double * koth_score, long comp_nr, int ksize);


void fitness_kothsimi(warrior_t * w, int anz, int update) {
    static double * koth_score=0;
    static warrior_t * koth_warriors=0;
    long ksize = globals.koth_size+1;
    static int worst_nr=0;
    int similar_nr, new_nr;
	int temp_score, temp_fights;
    int i,x,y,fights, old_fights;
    
    if (koth_warriors == 0) {
        koth_init(&koth_warriors, &koth_score, ksize);
        worst_nr = koth_worst_nr(koth_score, ksize);
    }
    
    switch(update) {    
    case 0:
        /* worst warrior is always replaced */
        for (i=0; i<anz; i++) {
            w[i].score=0;
            w[i].fights=0;

			fights=0;
            for (y=0; y<ksize; y++) {
                if (y == worst_nr) continue;

                /* calculate nr. of fights... */
				old_fights = fights;
                fights = (int)(((y+1.0) * globals.number_of_battles) / ksize);
				mars(&w[i], &koth_warriors[y], fights-old_fights);
            }
        }
        break;
        
    case 1:
        fights=globals.koth_battles;
        for (i=0; i<anz; i++) {
			new_nr = worst_nr;

            /* worst warrior is always replaced */
            koth_warriors[new_nr] = w[i];
            koth_warriors[new_nr].no=0;

			temp_score=0;
			temp_fights=0;


            for (y=0; y<ksize; y++) {
                if (y == new_nr) continue; /* no self-fights */

				koth_warriors[y].score=0; koth_warriors[y].fights=0;
                koth_warriors[y].no++;
	            koth_warriors[new_nr].score=0; koth_warriors[new_nr].fights=0;

                mars(&koth_warriors[new_nr], &koth_warriors[y], fights);

				temp_score += koth_warriors[new_nr].score;
				temp_fights += koth_warriors[new_nr].fights;
	
                koth_score[y + new_nr*ksize] = 100.0*koth_warriors[new_nr].score/koth_warriors[new_nr].fights;
                koth_score[new_nr + y*ksize] = 100.0*koth_warriors[y].score/koth_warriors[y].fights;

				koth_score[y + y*ksize]=0;
				for (x=0; x<ksize; x++) {
					if (x == y) continue;
					koth_score[y + y*ksize] += koth_score[x + y*ksize]/(ksize-1);
				}


            }

			koth_score[new_nr + new_nr*ksize] = 100.0*temp_score/temp_fights;

            similar_nr = get_most_similar_warrior(koth_score, new_nr, ksize);
            if (similar_nr != new_nr) {
                worst_nr = similar_nr;
                printf("Warrior %d updated with %d\n", worst_nr, new_nr);
                save_warrior(&koth_warriors[new_nr], new_nr, globals.all_battles, "kothsimi");
                print_kothsimi_status(koth_score, koth_warriors, ksize);
            }
            else {
                worst_nr = koth_worst_nr(koth_score, ksize);
				if (koth_score[worst_nr + worst_nr*ksize] * globals.koth_replace_treshold <
					koth_score[new_nr + new_nr*ksize]) {
					printf("Warrior replaced, nr %d\n", new_nr);
					save_warrior(&koth_warriors[new_nr], new_nr, globals.all_battles, "kothsimi");
					print_kothsimi_status(koth_score, koth_warriors, ksize);
				}
				else {
					worst_nr = new_nr;
				}
			}
        }
    }
}


int get_most_similar_warrior(double * koth_score, long comp_nr, int ksize) {
    int x, y, nr=0;
    double diff, replace_diff=0;
    double min_diff;

    min_diff = globals.koth_diff_treshold*globals.koth_diff_treshold;
    
    for (y=0; y<ksize; y++) {
        if (y == comp_nr) continue;
        
        diff = 0;
        for (x=0; x<ksize; x++) {
            if (x == y || x == comp_nr) continue;
			diff +=(((koth_score[x + y*ksize] - koth_score[y + y*ksize]) - (koth_score[x + comp_nr*ksize] - koth_score[comp_nr + comp_nr*ksize])) *
					((koth_score[x + y*ksize] - koth_score[y + y*ksize]) - (koth_score[x + comp_nr*ksize] - koth_score[comp_nr + comp_nr*ksize])));
        }
		diff /=(ksize-2);
        if ((diff < min_diff) && (koth_score[y + y*ksize] < koth_score[nr + nr*ksize])) {
			replace_diff = diff;
			nr = y;
        }
    }

	/* min_diff = sqrt(min_diff/(ksize-2)); */
    printf("diff: %6.2f, worst_score %5.2f, comp_scire %5.2f\n", replace_diff, koth_score[nr + nr*ksize], koth_score[comp_nr + comp_nr*ksize]);
	if (koth_score[nr + nr*ksize] < koth_score[comp_nr + comp_nr*ksize]) return nr;
    return comp_nr;
}




/* return nr of worst warrior */
int koth_worst_nr(double * koth_score, int size) {
	int y;
    int worst_nr=0;
    double worst_score=1e50;
        
    /* set nr of worst warrior */
	for (y=0; y<size; y++) {
		if (koth_score[y + y*size] < worst_score || y==0) {
            worst_nr = y;
            worst_score = koth_score[y + y*size];
        }
    }
    return worst_nr;
}



void koth_init(warrior_t ** koth_warriors, double ** koth_score, long ksize) {
    int x,y;
    warrior_t * w;
    double * score = *koth_score;
	int temp_score, temp_fights;
    
    *koth_warriors = malloc(sizeof(warrior_t)*ksize);
    *koth_score = malloc(sizeof(double)*ksize*ksize);
    w = *koth_warriors;
    score = *koth_score;
    
    create_warriors(w, ksize, globals.startlength, "kothsimi");
    
    /* recalc scores with no self-fights */
    printf("calculating hill-scores...\n");
    w[0].no=0; /* set Age of the warrior */
    
    for (y=1; y<ksize; y++) {
        w[y].no=0; /* set Age of the warrior */
		temp_score=0;
		temp_fights=0;
        for (x=0; x<y; x++) {
            printf("fight: %3d vs. %3d\n", y, x);
            w[x].score=0; w[x].fights=0;
            w[y].score=0; w[y].fights=0;
            
            mars(&w[x], &w[y], globals.koth_battles);

			temp_score += w[y].score;
			temp_fights += w[y].fights;

            score[x+y*ksize] = 100.0 * w[y].score / w[y].fights;
            score[y+x*ksize] = 100.0 * w[x].score / w[x].fights;
        }
		score[y + y*ksize] = 100.0*temp_score/temp_fights;
    }
}


void print_kothsimi_status(double * scores, warrior_t * w, int anz) {
    FILE * output;
    char file_name[256];
    int x, y, i, j;
    double score;
    warrior_t ** list=0;
    double max_color=255;
    int color;

    /* open status-file */
    sprintf(file_name, "%s%sstatus.htm", "kothsimi", globals.path_symbol);
    if (!(output = fopen(file_name, "w"))) {
        printf("Oh... I can not create the outputfile '%s'. No, don't panic.\n", file_name);
        printf("dont panic. Dont Panic! DONT PANIC!!! PAAANNNICCCC!\n");
        for (x=0; x<100; x++) printf(" PAAAAANIIIIIIIICCCC!!!");
        return;
    }

    /* calculate+set warrior scores */
    for (y=0; y<anz; y++) {
        w[y].score=0;
        for (x=0; x<anz; x++) {
            if (x == y) continue; /* no self-fights */
            w[y].score += scores[x + y*anz];
        }
        w[y].fights=anz-1;
    }

    /* get sorted list of warriors */
    list = sorted_list(w, anz);
    
    /* print header */
    fprintf(output, "<html><body bgcolor=\"#ffffff\"><center>\n");
    fprintf(output, "<h1>yace - king of the hill</h1>\n");
    fprintf(output, "<table cellpadding=\"2\" cellspacing=\"1\">\n<tr>");
    fprintf(output, "<td>&nbsp</td>");
    for (x=0; x<anz; x++)
        fprintf(output, "<td align=\"right\" bgcolor=\"#a4ddfe\"><a href=\"%08d.red\">%d</a></td>", list[anz-1-x] - &w[0], list[anz-1-x] - &w[0]);
    fprintf(output, "<td bgcolor=\"#a4ddfe\">Score</td><td bgcolor=\"#a4ddfe\">Age</td></tr>\n");
    for (i=0; i<anz; i++) {
        y = list[anz-1-i] - &w[0];

        /* <hr> for last warrior */
        if (i == anz-1) {
            fprintf(output, "<tr><td colspan=\"%d\"><hr noshade size=\"1\"></td></tr>\n", anz+3);
        }
        
        /* table row starts here */
        fprintf(output, "<tr><td align=\"right\" bgcolor=\"#a4ddfe\"><a href=\"%08d.red\">%d</a></td>", y, y);
        score=0;
        for (j=0; j<anz; j++) {
            x = list[anz-1-j] - &w[0];
            /* table cell starts here */
            if (x == y) { /* no self-fights */
                fprintf(output, "<td align=\"right\" bgcolor=\"#%02x%02x%02x\">&nbsp</td>", (int)(255-max_color/2), (int)(255-max_color/2), (int)(255-max_color/2));
                continue;
            }
            color = (int)(255.0 - max_color/300.0*scores[x + y*anz]);
            fprintf(output, "<td align=\"right\" bgcolor=\"#%02x%02x%02x\">", color, color, color);
            if (color <= 86) fprintf(output, "<font color=\"#fffff\">");
            fprintf(output, "%3.0f</td>", scores[x + y*anz]);
            if (color <= 86) fprintf(output, "</font>");
            score += scores[x + y*anz];
        }
        /* table row ends here */
        fprintf(output, "<td align=\"right\" bgcolor=\"dfdfdf\">%5.1f</td><td align=\"right\" bgcolor=\"dfdfdf\">%d</td></tr>\n",
                score/(anz-1), w[y].no);
    }
    
    fprintf(output, "</table></center>");
    fprintf(output, "<p>(c) by Martin Ankerl 2000");
    fprintf(output, "</body></html>");
    fclose(output);
}
