#ifndef ASM_H 
#define ASM_H
/* asm.h: token codes and macros for them. protos.
 * $Id: asm.h,v 1.1 2001/09/05 10:50:34 martin Exp $
 */

/* This file is part of `exhaust', a memory array redcode simulator.
 * Copyright (C) 2000 M Joonas Pihlaja
 *    
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * You may contact the author by:
 *   email: jpihlaja@cc.helsinki.fi
 *   mail:  M Joonas Pihlaja
 *          Artturi Kannistontie 4A11
 *          00320, Helsinki
 *          Finland
 */
#include <stdio.h>
/* macros:
 *  is_tok_opcode()     just what it says
 *  is_tok_pseudoop()   
 *  is_tok_modifier()   
 *  is_tok_mode()       is a token an addressing mode
 */

/*
 * asm_line() exit codes:
 *
 */

#define ASMLINE_ORG  -2     /* ORG encountered */
#define ASMLINE_DONE -1     /* END encountered */
#define ASMLINE_NONE  0     /* nothing to assemble, or something ignored */
#define ASMLINE_OK    1     /* assembled instruction OK */

/*
 * Token codes for the lexer.
 */
enum {
  TOK_FIRST = 256,      /* not used */

  TOK_OPCODE_START,
    TOK_DAT, TOK_SPL, TOK_MOV,  /* opcodes */
    TOK_JMP, TOK_JMZ, TOK_JMN, 
    TOK_ADD, TOK_SUB, TOK_SEQ, 
    TOK_SNE, TOK_MUL, TOK_DIV, 
    TOK_DJN, TOK_SLT, TOK_MOD,
    TOK_NOP,            /* 16 */
  TOK_OPCODE_STOP,

  TOK_PSEUDOOP_START,
    TOK_ORG,            /* pseudo-ops and start label */
    TOK_END,
    TOK_PIN,
  TOK_PSEUDOOP_STOP,
  
  TOK_MODIFIER_START,
    TOK_mF, TOK_mA, TOK_mB, /* modifiers */
    TOK_mAB, TOK_mBA, TOK_mX, TOK_mI,
  TOK_MODIFIER_STOP,

  TOK_MODE_START,
    TOK_DIRECT, TOK_IMMEDIATE,  /* addressing modes */
    TOK_BINDIRECT,  TOK_BPREDEC,
    TOK_BPOSTINC,  TOK_AINDIRECT,
    TOK_APREDEC,  TOK_APOSTINC, /* 8 */
  TOK_MODE_STOP,

  TOK_START,
  TOK_STR, TOK_INT,     /* string token, int token  */

  TOK_LAST          /* sentinel */
};


#define TOK_ERR TOK_FIRST   /* unused */

/*
 * Macros to identify a token type from a token code.
 */
#define is_tok_opcode(t)   ( t > TOK_OPCODE_START && t < TOK_OPCODE_STOP )
#define is_tok_pseudoop(t) ( t > TOK_PSEUDOOP_START && t < TOK_PSEUDOOP_STOP )
#define is_tok_modifier(t) ( t > TOK_MODIFIER_START && t < TOK_MODIFIER_STOP )
#define is_tok_mode(t)     ( t > TOK_MODE_START && t < TOK_MODE_STOP )


/*
 * protos
 */
extern int asm_line( char *line, insn_t *in, unsigned int CORESIZE );
extern void asm_file( FILE *F, warrior_t *w, unsigned int CORESIZE );
extern void asm_fname( const char *fname, warrior_t *w, unsigned int CORESIZE );
extern void dis1( char *s, insn_t in, unsigned int CORESIZE );
extern void discore( insn_t *core, int start, int end, unsigned int CORESIZE );


#endif /* ASM_H */
