#ifndef EXHAUST_H
#define EXHAUST_H
/*  exhaust.h:  Global constants, structures, and types
 * $Id: exhaust.h,v 1.1 2001/09/05 10:50:34 martin Exp $
 */

/* This file is part of `exhaust', a memory array redcode simulator.
 * Copyright (C) 2000 M Joonas Pihlaja
 *    
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * You may contact the author by:
 *   email: jpihlaja@cc.helsinki.fi
 *   mail:  M Joonas Pihlaja
 *          Artturi Kannistontie 4A11
 *          00320, Helsinki
 *          Finland
 */


/* global debug level */
#ifndef DEBUG
#define DEBUG 0
#endif

/*
 * Global constants
 *
 */

#define MAXLENGTH 100           /* max length of warrior */


/*
 * Global types
 *
 */

/* misc. integral types */
typedef unsigned char  u8_t;
typedef unsigned short u16_t;
typedef unsigned long  u32_t;
typedef long           s32_t;

  /* Choose an appropriate field_t width.  In a field_t variable
   * we hold unsigned integers in 0..CORESIZE-1
   */
#define FIELD_T_WIDTH 16

#if FIELD_T_WIDTH <= 8
  typedef u8_t field_t;
#elif FIELD_T_WIDTH <= 16
  typedef u16_t field_t;
#else
  typedef u32_t field_t;
#endif

  /*
   * Instructions in core:
   */
  typedef struct insn_st {
    u16_t in;                   /* flags, opcode, modifier, a- and b-modes */
    field_t a, b;               /* a-value, b-value */
  } insn_t;


  /*
   * Warrior struct
   */
typedef struct warrior_st {
    insn_t code[ MAXLENGTH ];   /* code of warrior */
    int len;                    /* length of -"- */
    int start;                  /* start relative to first insn */

    /* info fields -- these aren't automatically set or used */
    long score;
    long fights;
    int no;
} warrior_t;


#endif /* EXHAUST_H */
