#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "exhaust/insn.h"
#include "exhaust/exhaust.h"
#include "exhaust/sim.h"
#include "exhaust/asm.h"
#include "utils.h"
#include "../fitness/fitness.h"




/* numbertype 0: small number
   numbertype 1: big number
   numbertype 2: small or big number (random) */
int random_corenumber(int range, int numbertype) {
    int number = range-(random_number(range*2));

    /* random number mode */
    if (numbertype == 2) numbertype = random_number(2); 
    switch(numbertype) {
    case 0: break;
    case 1: number += globals.OPT_coresize/(random_number(3) + 2); break;
    }
    return (MODS(number, globals.OPT_coresize));
}

int random_number(int range) {
    int nr = rand()%range;
    if (nr < 0 || nr >= range)  {
        printf("random number error! got [%d], but range is [0, %d]!!!\n", nr, range-1);
    }
    return nr;
}

/********************************************/

int probability(double prop) {
    double nr;
    nr = (double)(rand()) / RAND_MAX;
    return (nr < prop);
}

void copy(const warrior_t * source, warrior_t *target) {
    *target = *source;
}
  

/********************************************/


void new_line_of_code(insn_t * codeline) {
    int op, m, ma, mb;

    codeline->a = random_corenumber(globals.max_number_size, 2);
    codeline->b = random_corenumber(globals.max_number_size, 2);
    
    op = random_number(TOK_OPCODE_STOP - TOK_OPCODE_START - 1);
    m =  random_number(TOK_MODIFIER_STOP - TOK_MODIFIER_START - 1);
    ma = random_number(TOK_MODE_STOP - TOK_MODE_START - 1);
    mb = random_number(TOK_MODE_STOP - TOK_MODE_START - 1);

    codeline->in = OP(op, m, ma, mb);
}

/********************************************/

int test_viable(warrior_t * test_warrior) {
    static warrior_t plant_warrior;
    static int plant_setup=0;

    /* set up plant */
    if (plant_setup == 0) {
        plant_warrior.code[0].in = OP(SPL, mF, DIRECT, DIRECT);
        plant_warrior.code[0].a = 0;
        plant_warrior.code[0].b = 0;
        plant_warrior.len = 1;
        plant_warrior.start=0;
        
        plant_setup=1;
    }
    
    
    /*test with plant*/
    test_warrior->score=0;
    test_warrior->fights=0;
    
    mars(&plant_warrior, test_warrior, 1);

    if (test_warrior->score == 0) return 0;
    return 1;

} /*test_viable()*/

/********************************************/

void create_warriors(warrior_t * warriors, int population_size, int linecount, char * directory) {
    FILE * warrior_file;
    int i;
    char file_name[256];

    /* test linecout */
    if (linecount > globals.max_instructions) linecount=globals.max_instructions;
    else if (linecount < 1) linecount=1;
    
    for (i=0; i < population_size; i++) {
        /* test if saved warrior exists */
        sprintf(file_name, "%s/%08d.red", directory, i);
	printf("[%s]\n", file_name);

        if (!(warrior_file = fopen(file_name, "r"))) {
            create_random_warrior(&warriors[i], linecount);
            printf("Warrior %5d created\n", i);
        }
        else {
                        fclose(warrior_file);
            /* load warrior from disk */
            asm_fname(file_name, &warriors[i], globals.OPT_coresize);
            printf("Warrior %5d loaded from %s\n", i, directory);
        }
        warriors[i].fights=0;
        warriors[i].score=0;
    }
}/*create_seed_population()*/

/********************************************/

void evolve_a_line(insn_t * codeline) {
    int o, m, ma, mb;
    
    /* opcode */
    if (probability(globals.mutation_rate)) o = random_number(TOK_OPCODE_STOP - TOK_OPCODE_START - 1);
    else o = (codeline->in >> opPOS) & opMASK;

    /* modifiers */
    if (probability(globals.mutation_rate))  m = random_number(TOK_MODIFIER_STOP - TOK_MODIFIER_START - 1);
    else m = (codeline->in >> moPOS) & moMASK;

    /* address mode a */
    if (probability(globals.mutation_rate)) ma = random_number(TOK_MODE_STOP - TOK_MODE_START - 1);
    else ma = (codeline->in >> maPOS) & mMASK;

    /* address mode b */
    if (probability(globals.mutation_rate)) mb = random_number(TOK_MODE_STOP - TOK_MODE_START - 1);
    else mb = (codeline->in >> mbPOS) & mMASK;
    
    /* number a */
    if (probability(globals.mutation_rate)) codeline->a = random_corenumber(globals.max_number_size, 2);

    /* number b */
    if (probability(globals.mutation_rate)) codeline->b = random_corenumber(globals.max_number_size, 2);

    /* create instruction */
    codeline->in = OP(o, m, ma, mb);
}/*evolve_a_line()*/

/********************************************/

char* decode_line(const insn_t * codeline) {
    static char * line_ptr=0;
    static char line[256];
    int h;

    line_ptr = line;
    
    h = ((codeline->in >> opPOS) & opMASK) + TOK_OPCODE_START + 1;
    switch (h) {
    case TOK_DAT: strncpy(line_ptr, "dat", 3); break;
    case TOK_SPL: strncpy(line_ptr, "spl", 3); break;
    case TOK_MOV: strncpy(line_ptr, "mov", 3); break;
    case TOK_JMP: strncpy(line_ptr, "jmp", 3); break;
    case TOK_JMZ: strncpy(line_ptr, "jmz", 3); break;
    case TOK_JMN: strncpy(line_ptr, "jmn", 3); break;
    case TOK_ADD: strncpy(line_ptr, "add", 3); break;
    case TOK_SUB: strncpy(line_ptr, "sub", 3); break;
    case TOK_SEQ: strncpy(line_ptr, "seq", 3); break;
    case TOK_SNE: strncpy(line_ptr, "sne", 3); break;
    case TOK_MUL: strncpy(line_ptr, "mul", 3); break;
    case TOK_DIV: strncpy(line_ptr, "div", 3); break;
    case TOK_DJN: strncpy(line_ptr, "djn", 3); break;
    case TOK_SLT: strncpy(line_ptr, "slt", 3); break;
    case TOK_MOD: strncpy(line_ptr, "mod", 3); break;
    case TOK_NOP: strncpy(line_ptr, "nop", 3); break;
    default: printf("ARRG!! opcode [%d]\n", h-(TOK_OPCODE_START + 1)); break;
    }

    line_ptr += 3;

    h = ((codeline->in >> moPOS) & moMASK) + TOK_MODIFIER_START + 1;
    switch(h) {
    case TOK_mF:  strncpy(line_ptr, ".f  ", 4); break;
    case TOK_mA:  strncpy(line_ptr, ".a  ", 4); break;
    case TOK_mB:  strncpy(line_ptr, ".b  ", 4); break;
    case TOK_mAB: strncpy(line_ptr, ".ab ", 4); break;
    case TOK_mBA: strncpy(line_ptr, ".ba ", 4); break;
    case TOK_mX:  strncpy(line_ptr, ".x  ", 4); break;
    case TOK_mI:  strncpy(line_ptr, ".i  ", 4); break;
    default: printf("ARRG!! modifier [%d]\n", h-(TOK_MODIFIER_START + 1)); break;
    }

    line_ptr += 4;

    h = ((codeline->in >> maPOS) & mMASK) + TOK_MODE_START + 1;
    switch(h) {
    case TOK_DIRECT:    strncpy(line_ptr, "$", 1); break;
    case TOK_IMMEDIATE: strncpy(line_ptr, "#", 1); break;
    case TOK_AINDIRECT: strncpy(line_ptr, "*", 1); break;
    case TOK_BINDIRECT: strncpy(line_ptr, "@", 1); break;
    case TOK_APREDEC:   strncpy(line_ptr, "{", 1); break;
    case TOK_BPREDEC:   strncpy(line_ptr, "<", 1); break;
    case TOK_APOSTINC:  strncpy(line_ptr, "}", 1); break;
    case TOK_BPOSTINC:  strncpy(line_ptr, ">", 1); break;
    default: printf("ARRG!! mode A [%d]\n", h-(TOK_MODE_START + 1)); break;
    }

    line_ptr += 1;
    
    sprintf(line_ptr,"%7d, ", codeline->a);
    
    line_ptr += 9;
    
    h = ((codeline->in >> mbPOS) & mMASK) + TOK_MODE_START + 1;
    switch(h) {
    case TOK_DIRECT:    strncpy(line_ptr, "$", 1); break;
    case TOK_IMMEDIATE: strncpy(line_ptr, "#", 1); break;
    case TOK_AINDIRECT: strncpy(line_ptr, "*", 1); break;
    case TOK_BINDIRECT: strncpy(line_ptr, "@", 1); break;
    case TOK_APREDEC:   strncpy(line_ptr, "{", 1); break;
    case TOK_BPREDEC:   strncpy(line_ptr, "<", 1); break;
    case TOK_APOSTINC:  strncpy(line_ptr, "}", 1); break;
    case TOK_BPOSTINC:  strncpy(line_ptr, ">", 1); break;
    default: printf("ARRG!! mode B [%d]\n", h-(TOK_MODE_START + 1)); break;
    }

    line_ptr += 1;

    sprintf(line_ptr,"%7d", codeline->b);
    
    line_ptr += 7;

    strcpy(line_ptr,"");

    return (line);
}
    
/********************************************/

void create_successor(const warrior_t * sw, warrior_t * tw) {
    int s_pos=0;
    int t_pos=0;

    if (sw == tw) {
        printf("Panic!! creat_successor called with Sourceaddress equal to Targetaddress!!\n");
        exit(1);
    }
    
    do {
        for (s_pos=0, t_pos=0; s_pos < sw->len; s_pos++, t_pos++) {
            if (t_pos == globals.max_instructions) {
                t_pos--;
                break;
            }
            if (probability(globals.insertion_rate)) {
                
                if(probability(globals.duplication_rate))
                    tw->code[t_pos]=sw->code[s_pos];
                else new_line_of_code(&tw->code[t_pos]);            
                s_pos--;
                continue;
            }
            if (probability(globals.removal_rate)) {
                t_pos--;
                continue;
            }
        
            /* sw should be copied!! (does this work? Debug needed!) */
            tw->code[t_pos] = sw->code[s_pos];
            evolve_a_line(&tw->code[t_pos]);
        }
        tw->len = t_pos;
        tw->start = 0;
        tw->score=0;
        tw->fights=0;
    } while (tw->len > globals.max_instructions || test_viable(tw)==0);

#ifdef DEBUG_YACE
    for (i=0; i < tw->len; i++) {
        printf("%s\n", decode_line(&tw->code[i]));
    }
    printf("\n");
#endif     
}/*create_successor()*/

/********************************************/

void straight_replace(const warrior_t * sw, warrior_t * tw) {
    *tw = *sw;
}

/********************************************/

void mars(warrior_t *war1, warrior_t *war2, int rounds) {
    int i=0, j=0, pos=0;
    int result=0;
    int start1=0;
    int start2=0;
        
    for (i=0; i<rounds; i++)
    {
        pos = globals.OPT_minsep + random_number(globals.OPT_coresize - globals.OPT_minsep*2);
                
        for (j=0; j<2; j++)
        {
            /* coreclear */
            memset( core, 0, sizeof(insn_t) * globals.OPT_coresize );

            start1=0;
            start2=0;
            if (j)  start1 = pos;
            else    start2 = pos;
                        
            memcpy(core + start1, war1->code, sizeof(insn_t) * war1->len);
            memcpy(core + start2, war2->code, sizeof(insn_t) * war2->len);
            result = sim( 2, war1->start + start1, war2->start + start2, globals.OPT_cycle, NULL );

            /* one-on-one two warriors:
               0: warrior 1 won
               1: warrior 2 won
               2: tie
               -1: simulator panic attack -- something's gone wrong */
            switch(result)
            {
            case 0: war1->score += 3; break;
            case 1: war2->score += 3; break;
            case 2: war1->score += 1; war2->score += 1; break;
            case -1: printf("simulator panic attack!\n"); exit(1); break;
            }
            war1->fights += 1;
            war2->fights += 1;
            globals.all_battles++;
        }/*for(0 --> 1)*/
    }/*for (0 --> rounds)*/
}/*mars()*/

/********************************************/

void status(warrior_t * warriors, int anz, int current_cycle) {
    double bestscore_adjusted;
    int nr;

    /* calculate fitness values */
    printf("Status update. %ld fights already.\n", globals.all_battles);
    fitness_testset(warriors, anz, 0);
    printf(" Scores:\n");

    /* print scores, get nr of best warrior */
    nr = print_scores(warriors, anz);

    /* calculate best warrior more precise */
    fitness_testset(&warriors[nr], 1, 1);
    bestscore_adjusted = 100.0 * warriors[nr].score / warriors[nr].fights;
    
    printf("\n");
    printf("Best one: nr. %d, fights %ld, score %5.1f.\n", nr+1, warriors[nr].fights, bestscore_adjusted);
    
    /* save best warrior */
    save_warrior(&warriors[nr], (int)(bestscore_adjusted * 1E5), current_cycle, globals.dir_status);

    printf("Status update finished.\n");
}

void save_warrior(const warrior_t * warrior, unsigned long save_nr, int nr, char * directory) {
    char file_name[256];
    FILE * warrior_file=0;
    int i;
    
    sprintf(file_name, "%s%s%08ld.red", directory, globals.path_symbol, save_nr);
    if (0==(warrior_file=fopen(file_name,"w+"))) {
        printf("Can not write [%s] to disk!\n", file_name);
        exit(1);
    }
    
    fprintf(warrior_file, ";redcode-94\n");
    fprintf(warrior_file, ";assert 1\n");
    fprintf(warrior_file, ";author Martin Ankerl <martinankerl@web.de>\n");
    fprintf(warrior_file, ";name Evolver %5d x %5ld\n", nr, globals.population_size);
    fprintf(warrior_file, ";strategy created with YACE - yet another corewar evolver\n");
    fprintf(warrior_file, ";strategy http://www.fhs-hagenberg.ac.at/students/se/se00001/\n");
    
    for (i=0; i < warrior->len; i++) {
        fprintf(warrior_file, "%s\n", decode_line(&warrior->code[i]));
    }
    fprintf(warrior_file, "end 0\n");
    fclose(warrior_file);
}

    


void save_warriors(const warrior_t * warriors, int anz, int current_cycle, char * directory) {
    int i;

    printf("Saving %s to disk\n", directory);
    for (i=0; i<anz; i++) {
        /* print warriors to file */
        save_warrior(&warriors[i], i, current_cycle, directory);
        
    }
}




void setup(warrior_t ** warriors, warrior_select_t ** wsel) {
    FILE * del2stop;
    FILE * config_file=0;
    char buffer[256];
        
    /*seed the random numbers*/
    /* srand(123); */
    time_t t;       
    int nr;    
    nr = (unsigned int) time(& t);
    printf("srand [%d]\n", nr);
    srand(nr);

    /* all params are set to -1 to enshure everything is specified in
       configfile */
    globals.population_size=-1;
    globals.number_of_battles=-1;
    globals.OPT_coresize=-1;
    globals.OPT_cycle=-1;
    globals.OPT_minsep=-1;
    globals.OPT_processes=-1;
    globals.mutation_rate=-1;
    globals.insertion_rate=-1;
    globals.removal_rate=-1;
    globals.duplication_rate=-1;
    globals.reintroduce_veteran_rate=-1;
    globals.max_number_size=-1;
    globals.max_instructions=-1;
    globals.startlength=-1;
    globals.testset_size=-1;
    globals.testset_battles=-1;
    globals.koth_size=-1;
    globals.koth_battles=-1;
    globals.cycles_before_save=-1;
    globals.cycles_before_status=-1;
    globals.koth_replace_treshold=-1;
    globals.koth_diff_treshold=-1;
    globals.evolver_fitness=-1;
    globals.evolver_replacement=-1;
    strcpy(globals.path_symbol,"x");
    strcpy(globals.dir_status,"status");


    /* these variables are not set */
    globals.all_battles=0;

    /*read config file if there*/
    if (0==(config_file=fopen("yace.cfg","r"))) {
        printf("yace.cfg not found!\n");
        exit(1);
    }
    else {
        while (fgets(buffer,255,config_file)!=0) {                
            /*ignore line beginning with ';' */
            if (buffer[0] == ';') continue;
            if (buffer2long(buffer, "population_size", &globals.population_size)) continue;
            if (buffer2long(buffer, "max_instructions", &globals.max_instructions)) continue;
            if (buffer2long(buffer, "max_number_size", &globals.max_number_size)) continue;
            if (buffer2long(buffer, "startlength", &globals.startlength)) continue;

            if (buffer2double(buffer, "mutation_rate", &globals.mutation_rate)) continue;
            if (buffer2double(buffer, "insertion_rate", &globals.insertion_rate)) continue;
            if (buffer2double(buffer, "removal_rate", &globals.removal_rate)) continue;
            if (buffer2double(buffer, "duplication_rate", &globals.duplication_rate)) continue;
            if (buffer2double(buffer, "reintroduce_veteran_rate", &globals.reintroduce_veteran_rate)) continue;

            if (buffer2long(buffer, "number_of_battles", &globals.number_of_battles)) continue;
            
            if (buffer2long(buffer, "OPT_coresize", &globals.OPT_coresize)) continue;
            if (buffer2long(buffer, "OPT_cycle", &globals.OPT_cycle)) continue;
            if (buffer2long(buffer, "OPT_minsep", &globals.OPT_minsep)) continue;
            if (buffer2long(buffer, "OPT_processes", &globals.OPT_processes)) continue;
            
            if (buffer2long(buffer, "testset_size", &globals.testset_size)) continue;
            if (buffer2long(buffer, "testset_battles", &globals.testset_battles)) continue;
                
            if (buffer2long(buffer, "koth_size", &globals.koth_size)) continue;
            if (buffer2long(buffer, "koth_battles", &globals.koth_battles)) continue;
                        
            if (buffer2double(buffer, "koth_replace_treshold", &globals.koth_replace_treshold)) continue;
            if (buffer2double(buffer, "koth_diff_treshold", &globals.koth_diff_treshold)) continue;

            if (buffer2long(buffer, "cycles_before_status", &globals.cycles_before_status)) continue;
            if (buffer2long(buffer, "cycles_before_save", &globals.cycles_before_save)) continue;

            if (buffer2long(buffer, "evolver_fitness", &globals.evolver_fitness)) continue;
            if (buffer2long(buffer, "evolver_replacement", &globals.evolver_replacement)) continue;

            if (buffer2string(buffer, "path_symbol", globals.path_symbol)) continue;
            if (buffer2string(buffer, "status", globals.dir_status)) continue;
        }/*while*/
        fclose(config_file);
    }
    
    /* popsize should be dividable by 2! */
    if (globals.population_size % 2) globals.population_size++;

    /* allocate whole memory */
    *warriors = malloc(sizeof(warrior_t)*globals.population_size);
    *wsel     = malloc(sizeof(warrior_select_t)*globals.population_size + 1);
    
    /* request address of core memory */
    core = sim_alloc_bufs( 2, globals.OPT_coresize, globals.OPT_processes, globals.OPT_cycle);

    /* test if all parameters have been set */
    if(globals.population_size == -1) { printf("globals.population_size not set.\n"); exit(1); }
    if(globals.number_of_battles == -1) { printf("globals.number_of_battles not set.\n"); exit(1); }    
    if(globals.OPT_coresize == -1) { printf("globals.OPT_coresize not set.\n"); exit(1); }  
    if(globals.OPT_cycle == -1) { printf("globals.OPT_cycle not set.\n"); exit(1); }    
    if(globals.OPT_minsep == -1) { printf("globals.OPT_minsep not set.\n"); exit(1); }  
    if(globals.OPT_processes == -1) { printf("globals.OPT_processes not set.\n"); exit(1); }    
    if(globals.mutation_rate == -1) { printf("globals.mutation_rate not set.\n"); exit(1); }    
    if(globals.insertion_rate == -1) { printf("globals.insertion_rate not set.\n"); exit(1); }  
    if(globals.removal_rate == -1) { printf("globals.removal_rate not set.\n"); exit(1); }  
    if(globals.duplication_rate == -1) { printf("globals.duplication_rate not set.\n"); exit(1); }
    if(globals.reintroduce_veteran_rate == -1) { printf("globals.reintroduce_veteran_rate not set.\n"); exit(1); }
    if(globals.max_number_size == -1) { printf("globals.max_number_size not set.\n"); exit(1); }    
    if(globals.max_instructions == -1) { printf("globals.max_instructions not set.\n"); exit(1); }  
    if(globals.startlength == -1) { printf("globals.startlength not set.\n"); exit(1); }    
    if(globals.testset_size == -1) { printf("globals.testset_size not set.\n"); exit(1); }  
    if(globals.testset_battles == -1) { printf("globals.testset_battles not set.\n"); exit(1); }  
    if(globals.koth_size == -1) { printf("globals.koth_size not set.\n"); exit(1); }    
    if(globals.koth_battles == -1) { printf("globals.koth_battles not set.\n"); exit(1); }  
    if(globals.cycles_before_save == -1) { printf("globals.cycles_before_save not set.\n"); exit(1); }  
    if(globals.cycles_before_status == -1) { printf("globals.cycles_before_status not set.\n"); exit(1); }  
    if(globals.path_symbol[0] == 'x') { printf("globals.path_symbol not set.\n"); exit(1); }
    if(globals.evolver_fitness == -1) { printf("globals.evolver_fitness not set.\n"); exit(1); }  
    if(globals.evolver_replacement == -1) { printf("globals.evolver_replacement not set.\n"); exit(1); }  
    if(globals.koth_replace_treshold == -1) { printf("globals.koth_replace_treshold not set.\n"); exit(1); }  
    if(globals.koth_diff_treshold == -1) { printf("globals.koth_diff_treshold not set.\n"); exit(1); }  

    /* create del2stop */
    if (!(del2stop = fopen("del2stop", "w"))) {
        printf("Can not open 'del2stop' for writing!");
        exit(1);
    }
    else {
        fprintf(del2stop, "If you want to stop the evolver, delete\n");
        fprintf(del2stop, "this file and yace will quit after the\n");
        fprintf(del2stop, "current cycle is finished.\n");
        fclose(del2stop);
    }

    /* print configuration */
    printf("globals.population_size=[%ld]\n", globals.population_size);
    printf("globals.number_of_battles=[%ld]\n", globals.number_of_battles);
    printf("globals.OPT_coresize=[%ld]\n", globals.OPT_coresize);
    printf("globals.OPT_cycle=[%ld]\n", globals.OPT_cycle);
    printf("globals.OPT_minsep=[%ld]\n", globals.OPT_minsep);
    printf("globals.OPT_processes=[%ld]\n", globals.OPT_processes);
    printf("globals.mutation_rate=[%f]\n", globals.mutation_rate);
    printf("globals.insertion_rate=[%f]\n", globals.insertion_rate);
    printf("globals.removal_rate=[%f]\n", globals.removal_rate);
    printf("globals.duplication_rate=[%f]\n", globals.duplication_rate);
    printf("globals.reintroduce_veteran_rate=[%f]\n", globals.reintroduce_veteran_rate);
    printf("globals.max_number_size=[%ld]\n", globals.max_number_size);
    printf("globals.max_instructions=[%ld]\n", globals.max_instructions);
    printf("globals.startlength=[%ld]\n", globals.startlength);
    printf("globals.testset_size=[%ld]\n", globals.testset_size);
    printf("globals.testset_battles=[%ld]\n", globals.testset_battles);
    printf("globals.koth_size=[%ld]\n", globals.koth_size);
    printf("globals.koth_battles=[%ld]\n", globals.koth_battles);
    printf("globals.cycles_before_save=[%ld]\n", globals.cycles_before_save);
    printf("globals.cycles_before_status=[%ld]\n", globals.cycles_before_status);
    printf("globals.koth_replace_treshold=[%f]\n", globals.koth_replace_treshold);
    printf("globals.koth_diff_treshold=[%f]\n", globals.koth_diff_treshold);
    printf("globals.path_symbol=[%s]\n", globals.path_symbol);
    printf("globals.evolver_fitness=[%ld]\n", globals.evolver_fitness);
    printf("globals.evolver_replacement=[%ld]\n", globals.evolver_replacement);
        

    /* these variables are not set */
    printf("globals.all_battles=[%ld]\n", globals.all_battles);

}


/********************************************/

int buffer2double(const char* buffer, const char* searchstring, double *number) {
    if(strstr(buffer, searchstring)!=0)     {
        char * parm_ptr=strstr(buffer,"=");
        if (parm_ptr!=0) {
            *number=atof(++parm_ptr);
            return 1;
        }
    }
    return 0;
}

int buffer2long(const char* buffer, const char* searchstring, long *number) {
    if(strstr(buffer, searchstring)!=0)     {
        char * parm_ptr=strstr(buffer,"=");
        if (parm_ptr!=0)  {
            *number=atol(++parm_ptr);
            return 1;
        }
    }
    return 0;
}

int buffer2string(const char* buffer, const char* searchstring, char* target) {
    char* parm_ptr=strstr(buffer,"=");
    char* temp=target;
    if(!strstr(buffer, searchstring)) return 0;

    if (!parm_ptr) return 0;
    parm_ptr++;
    
    /* copy string */
    while(*temp++ = *parm_ptr++);
    
    /* remove '\n' */
    parm_ptr = strchr(target,'\n');
    if (parm_ptr != NULL) *parm_ptr = '\0';
    return 1;
}

/* returns a pointer to a random list */
int * random_list(int number) {
    int i;
    int nr;
    static int old_size=0;
    static int * list=0;

    /* realloc if necessery */
    if (old_size < number) {
        if (old_size > 0) free (list);
        list = malloc (sizeof(int)*number);
        old_size=number;
    }
    
    /* set numbers */
    for (i=0; i<number; i++) list[i] = i;
    
    /* create random numbers, every number only once */
    for (i=0; i<number-1; i++) { /* number-1: number would cause an infinite loop */
        if (list[i] == i) {
            nr = random_number(number-i-1)+i+1;
            list[i] = list[nr];
            list[nr] = i;
        }
    }

    /*check last number */
    if (list[number-1] == number-1) {
        nr = random_number(number-1); /* replace with any number except last one (itself) */
        list[number-1] = list[nr];
        list[nr]       = number-1;
    }
    return list;
}

int compare_warriors(const void *a, const void *b) {
    double sa=0, sb=0;
    
    warrior_t * wa = *(warrior_t **)a;
    warrior_t * wb = *(warrior_t **)b;
    
    sa = 1.0 * wa->score / wa->fights;
    sb = 1.0 * wb->score / wb->fights;
    if (sa>sb) return 1;
    if (sa<sb) return -1;
    return 0;
}


/* returns a pointer to a score-sorted list of pointers. best individual fist. */
warrior_t ** sorted_list(warrior_t * w, int number) {
    int i;
    static int old_size=0;
    static warrior_t ** list=0;
    
    /* realloc only when necessery */
    if (old_size < number) {
        if (old_size > 0) free (list);
        list = malloc(sizeof(warrior_t*)*number);
        old_size=number;
    }
    
    /* set pointers */
    for (i=0; i<number; i++) list[i] = &w[i];
    
    qsort(list, number, sizeof(warrior_t*), compare_warriors);
    
    return list;
}

int shutdowntest() {
    FILE * del2stop;
    
    if (!(del2stop = fopen("del2stop", "r"))) {
        printf("Shutting down sequence initiated...\n");
        return 1;
    }
    fclose (del2stop);
    return 0;
}


void create_random_warrior(warrior_t * warrior, int linecount) {
    int i;
    do {
        for (i=0; i<linecount; i++) {
            new_line_of_code(&warrior->code[i]);
        }
        warrior->len = linecount;
        warrior->start = 0;
    } while (!test_viable(warrior));
    warrior->fights=0;
    warrior->score=0;
}


int get_best_warrior(warrior_t * w, int anz) {
    int nr;
    int best_nr=0;
    double best_score=-1;
    double score=0;
    
    for (nr=0; nr<anz; nr++) {
        score = 100.0 * w[nr].score / w[nr].fights;
        if (score > best_score) {
            best_score = score;
            best_nr = nr;
        }
    }
    return best_nr;
}


/* returns best warrior */
int print_scores(warrior_t * warriors, int anz) {
    int i;
    double score_adjusted=0;
    double bestscore_adjusted=0;
    int nr=0;
    
    for (i=0; i<anz; i++) {
        score_adjusted = 100.0 * warriors[i].score / warriors[i].fights;
        printf("%5.1f  ", score_adjusted);
        if (!((i+1) % 10)) printf("\n");
        if (score_adjusted > bestscore_adjusted) {
            nr = i;
            bestscore_adjusted = score_adjusted;
        }
    }
    return nr;
}


/* evaluates how many fights are possible within 10 seconds */    
void benchmark() {
    const anz=10;
    const benchtime=15;
    int i, j;
    unsigned long nr;
    unsigned long times[4];
    unsigned long max, min;
    time_t  start, end, t;
    warrior_t * benchwar;
    
    /* allocate Memory for 10 benchmark-warriors */
    benchwar = malloc(sizeof(warrior_t)*anz);
    
    /* load benchmarkwarriors */
    printf("Starting benchmark, please wait...\n");

    /* init */
    nr = (unsigned int) time(& t);
    printf("srand [%d]\n", nr);
    srand(12);

    /* all params are set to -1 to enshure everything is specified in
       configfile */
    globals.population_size=10;
    globals.number_of_battles=30;
    globals.OPT_coresize=800;
    globals.OPT_cycle=8000;
    globals.OPT_minsep=20;
    globals.OPT_processes=800;
    globals.max_instructions=20;
    globals.max_number_size=20;
    strcpy(globals.path_symbol,"\\");
    /* request address of core memory */
    core = sim_alloc_bufs( 2, globals.OPT_coresize, globals.OPT_processes, globals.OPT_cycle);
    create_warriors(benchwar, anz, globals.startlength, "bench");

    printf("\nTo get good results, you should close all unneeded applications");
    printf("\nbefore running this benchmark.");
    printf("\n\nThis Benchmark calculates how many fights are possible within 20 seconds.");
    printf("\nIt is run 4 times to find out how much the results can differ.");
    printf("\n\nPlease tell me your System configuration and your score if you want to be added!");
    printf("\nto the top-list on http://www.fhs-hagenberg.ac.at/students/se/se00001/.");
    start = time(NULL);

    /*everyone against everyone */
    nr = 0;

    start = time(NULL)+1;
    while(time(NULL) < start);
    printf("\n\nbenchmarking, please wait approx. 1 minutes and 20 seconds.. \n\n");
    
    /* benchmark 4 times and calculate average */
    for (t=0; t<4; t++) {
        i=0;
        j=0;
        nr=0;
        
        /* wait till second starts */
        start = time(NULL)+1;
        while(time(NULL) < start);

        while(time(NULL)-start < benchtime) {
            if (i < anz-1) i++;
            else {
                i=0;
                if (j < anz-1) j++;
                else j=0;
            }
	    
            if (i - j) {
                mars(&benchwar[i], &benchwar[j], 1);
                nr+=2;
            }
        }
        times[t]=nr;
	printf("%d / 4: %.0f fights per second\n", t+1, times[t]/15.0);
    }
    max = times[0];
    for (t=1; t<4; t++) if (times[t] > max) max=times[t];
    min = times[0];
    for (t=1; t<4; t++) if (times[t] < min) min=times[t];

    printf("\n\n");
    printf("-------------------------------------------------------------\n");
    printf("Fights per second (the higher the better):           %7.2f\n", 1.0* (times[0]+times[1]+times[2]+times[3])/(benchtime*4));
    printf("Tolerance (should be lower than 2 percent):          %7.2f%%\n", 100.0*(max-min)/min);
    printf("-------------------------------------------------------------\n");
    printf("%4.2f times faster than AMD Duron650, PC133, Win2k.\n", 1.0*(times[0]+times[1]+times[2]+times[3])/(benchtime*4)/1348.27);
}


void Intro() {
    printf("YACE - yet another corewar evolver      V. 1.0\n");
    printf("usage: yace -h      this screen\n");
    printf("       yace -bench  starts a cool benchmark :-)\n");
    printf("       yace         starts evolving til file del2stop is deleted\n");
    printf("\n");
}

    
    
