#ifndef UTILS_H
#define UTILS_H

/* these are set from the config file*/

typedef struct globals_st {
    long population_size;
    long max_instructions;
    long max_number_size;
    long startlength;

    double mutation_rate;
    double insertion_rate;
    double removal_rate;
    double duplication_rate;
    double reintroduce_veteran_rate;
    long number_of_battles;

    long OPT_coresize;
    long OPT_cycle;
    long OPT_minsep;
    long OPT_processes;

    long testset_size;
    long testset_battles;
	
    long koth_size;
    long koth_battles;

    double koth_replace_treshold;
    double koth_diff_treshold;
	
    long cycles_before_save;
    long cycles_before_status;

    long evolver_fitness;
    long evolver_replacement;

    long all_battles;

    char path_symbol[32];
    char dir_status[256];

} globals_t;

globals_t globals;

typedef struct warrior_select_st {
    warrior_t * s[5];
} warrior_select_t;
    
    
insn_t *core; /* address of core */

/* prototypes *******/

void setup(warrior_t ** warriors, warrior_select_t ** wsel);

int buffer2double(const char* buffer, const char* searchstring, double *number);
int buffer2long(const char* buffer, const char* searchstring, long *number);
int buffer2string(const char* buffer, const char* searchstring, char* target);
int * random_list(int number);
char * decode_line(const insn_t * codeline);
void mars(warrior_t *war1, warrior_t *war2, int rounds);
void create_warriors(warrior_t * warriors, int population_size, int linecount, char * directory);
void reproduce_standard(const int * warriors_selection, warrior_t * warriors, int anz);
void status(warrior_t * warriors, int anz, int current_cycle);
void save_warriors(const warrior_t * warriors, int anz, int current_cycle, char * directory);
void straight_replace(const warrior_t * sw, warrior_t * tw);
void create_successor(const warrior_t * sw, warrior_t * tw);
int test_viable(warrior_t * test_warrior);
int probability(double prop);
int random_corenumber(int range, int numbertype);
int random_number(int range);
void new_line_of_code(insn_t * codeline);
void evolve_a_line(insn_t * codeline);
warrior_t ** sorted_list(warrior_t * w, int number);
int compare_warriors(const void *a, const void *b);
int shutdowntest();
void create_random_warrior(warrior_t * warrior, int linecount);
int get_best_warrior(warrior_t * w, int anz);
int print_scores(warrior_t * warriors, int anz);
void save_warrior(const warrior_t * warrior, unsigned long save_nr, int nr, char * directory);
void benchmark();
void Intro();


#endif
